package io.gitlab.jfronny.inceptum.launcher.system.export;

import io.gitlab.jfronny.commons.io.JFiles;
import io.gitlab.jfronny.inceptum.common.ConfigHolder;
import io.gitlab.jfronny.inceptum.common.Utils;
import io.gitlab.jfronny.inceptum.launcher.model.curseforge.CurseforgeModpackManifest;
import io.gitlab.jfronny.inceptum.launcher.model.inceptum.InstanceMeta;
import io.gitlab.jfronny.inceptum.launcher.model.multimc.MMCPackMeta;
import io.gitlab.jfronny.inceptum.launcher.util.ModPath;
import io.gitlab.jfronny.inceptum.launcher.util.ProcessState;
import io.gitlab.jfronny.inceptum.launcher.util.ignore.IgnoringWalk;
import io.gitlab.jfronny.inceptum.launcher.system.mds.IWModDescription;
import io.gitlab.jfronny.inceptum.launcher.system.mds.ModsDirScanner;
import io.gitlab.jfronny.inceptum.launcher.system.source.CurseforgeModSource;
import io.gitlab.jfronny.inceptum.launcher.system.source.ModSource;

import java.io.IOException;
import java.net.URISyntaxException;
import java.nio.file.*;
import java.time.Instant;
import java.util.*;

//TODO deduplicate
//TODO modrinth export

public class InstanceExporter {
    public static final int STEP_COUNT = 3;
    private static final String OVERRIDES_DIR_DEFAULT = "overrides";
    private static final String DOT_MINECRAFT = ".minecraft";

    public static void exportCurseZip(ProcessState state, Path instanceDir, InstanceMeta meta, ModsDirScanner mds, Path exportPath, String version) throws IOException, URISyntaxException {
        if (Files.exists(exportPath)) Files.delete(exportPath);
        try (FileSystem fs = Utils.openZipFile(exportPath, true)) {
            state.incrementStep("Preparing basic manifest");
            CurseforgeModpackManifest manifest = new CurseforgeModpackManifest();
            manifest.minecraft = new CurseforgeModpackManifest.Minecraft();
            manifest.minecraft.version = meta.getMinecraftVersion();
            manifest.minecraft.modLoaders = new LinkedHashSet<>();
            if (meta.isFabric()) {
                CurseforgeModpackManifest.Minecraft.ModLoader loader = new CurseforgeModpackManifest.Minecraft.ModLoader();
                loader.id = "fabric-" + meta.getLoaderVersion();
                loader.primary = true;
                manifest.minecraft.modLoaders.add(loader);
            }
            manifest.manifestType = "minecraftModpack";
            manifest.manifestVersion = 1;
            manifest.name = instanceDir.getFileName().toString();
            manifest.version = version;
            manifest.author = ConfigHolder.CONFIG.authorName;
            manifest.overrides = OVERRIDES_DIR_DEFAULT;
            Path overrides = fs.getPath(OVERRIDES_DIR_DEFAULT);
            Files.createDirectories(overrides);
            if (meta.isFabric()) {
                state.incrementStep("Adding mods");
                manifest.files = new LinkedHashSet<>();
                if (!mds.isComplete()) throw new IOException("Mods dir scan is not yet completed");
                modsLoop:
                for (IWModDescription mod : mds.getMods()) {
                    if (!ModPath.isEnabled(mod.path())) continue;
                    state.updateStep(mod.path().toString());
                    if (ModPath.isImod(mod.path())) {
                        Set<ModSource> sources = mod.mod().orElseThrow().sources.keySet();
                        for (ModSource source : sources) {
                            if (source instanceof CurseforgeModSource cms) {
                                CurseforgeModpackManifest.File f = new CurseforgeModpackManifest.File();
                                f.projectID = cms.getProjectId();
                                f.fileID = cms.getFileId();
                                f.required = true;
                                manifest.files.add(f);
                                continue modsLoop;
                            }
                        }
                        // Not available on CF
                        for (ModSource source : sources) {
                            Path jarPath = source.getJarPath();
                            Files.copy(jarPath, jarPath.resolve(mod.path().getFileName().toString()));
                            continue modsLoop;
                        }
                    }
                    Path md = overrides.resolve("mods");
                    Files.createDirectories(md);
                    Files.copy(mod.path(), md.resolve(mod.path().getFileName().toString()));
                }
            }
            state.incrementStep("Adding files");
            for (Path l : IgnoringWalk.walk(instanceDir).toList()) {
                String fn = l.getFileName().toString();
                if (fn.equals("mods") || fn.startsWith(".")) continue;
                state.updateStep(fn);
                Path target = overrides.resolve(fn);
                JFiles.copyContent(l, target);
            }
            JFiles.writeObject(fs.getPath("manifest.json"), manifest);
        }
    }

    public static void exportMultiMCZip(ProcessState state, Path instanceDir, InstanceMeta meta, ModsDirScanner mds, Path exportPath) throws IOException, URISyntaxException {
        if (Files.exists(exportPath)) Files.delete(exportPath);
        try (FileSystem fs = Utils.openZipFile(exportPath, true)) {
            {
                state.incrementStep("Preparing basic manifest");
                MMCPackMeta manifest = new MMCPackMeta();
                manifest.formatVersion = 1;
                manifest.components = new ArrayList<>();
                MMCPackMeta.Component lwjgl = new MMCPackMeta.Component();
                lwjgl.dependencyOnly = true;
                lwjgl.uid = "org.lwjgl3";
                lwjgl.version = "3.2.2"; //TODO get automatically
                manifest.components.add(lwjgl);
                MMCPackMeta.Component minecraft = new MMCPackMeta.Component();
                minecraft.important = true;
                minecraft.uid = "net.minecraft";
                minecraft.version = meta.getMinecraftVersion();
                manifest.components.add(minecraft);
                if (meta.isFabric()) {
                    MMCPackMeta.Component intermediary = new MMCPackMeta.Component();
                    intermediary.dependencyOnly = true;
                    intermediary.uid = "net.fabricmc.intermediary";
                    intermediary.version = meta.getMinecraftVersion();
                    manifest.components.add(intermediary);
                    MMCPackMeta.Component fabric = new MMCPackMeta.Component();
                    fabric.uid = "net.fabricmc.fabric-loader";
                    fabric.version = meta.getLoaderVersion();
                    manifest.components.add(fabric);
                }
                JFiles.writeObject(fs.getPath("mmc-pack.json"), manifest);
            }
            {
                state.updateStep("Preparing instance config");
                Files.writeString(fs.getPath("instance.cfg"), String.format("""
                        ForgeVersion=
                        InstanceType=OneSix
                        IntendedVersion=
                        JoinServerOnLaunch=false
                        LWJGLVersion=
                        LiteloaderVersion=
                        LogPrePostOutput=true
                        MCLaunchMethod=LauncherPart
                        OverrideCommands=false
                        OverrideConsole=false
                        OverrideGameTime=false
                        OverrideJavaArgs=false
                        OverrideJavaLocation=false
                        OverrideMCLaunchMethod=false
                        OverrideMemory=false
                        OverrideNativeWorkarounds=false
                        OverrideWindow=false
                        iconKey=default
                        lastLaunchTime=%s
                        lastTimePlayed=0
                        name=%s
                        notes=
                        totalTimePlayed=0
                        """, Instant.now().toEpochMilli(), instanceDir.getFileName()));
            }
            Path overrides = fs.getPath(DOT_MINECRAFT);
            Files.createDirectories(overrides);
            if (meta.isFabric()) {
                state.incrementStep("Adding mods");
                if (!mds.isComplete()) throw new IOException("Mods dir scan is not yet completed");
                modsLoop:
                for (IWModDescription mod : mds.getMods()) {
                    if (!ModPath.isEnabled(mod.path())) continue;
                    state.updateStep(mod.path().toString());
                    if (ModPath.isImod(mod.path())) {
                        Set<ModSource> sources = mod.mod().orElseThrow().sources.keySet();
                        for (ModSource source : sources) {
                            Path jarPath = source.getJarPath();
                            Files.copy(jarPath, jarPath.resolve(mod.path().getFileName().toString()));
                            continue modsLoop;
                        }
                    }
                    Path md = overrides.resolve("mods");
                    Files.createDirectories(md);
                    Files.copy(mod.path(), md.resolve(mod.path().getFileName().toString()));
                }
            }
            state.incrementStep("Adding files");
            for (Path l : IgnoringWalk.walk(instanceDir).toList()) {
                String fn = l.getFileName().toString();
                if (fn.equals("mods") || fn.startsWith(".")) continue;
                state.updateStep(fn);
                Path target = overrides.resolve(fn);
                JFiles.copyContent(l, target);
            }
        }
    }
}
