package io.gitlab.jfronny.inceptum;

import io.gitlab.jfronny.commons.io.JFiles;
import org.jetbrains.annotations.Nullable;

import java.io.Closeable;
import java.io.IOException;
import java.net.*;
import java.nio.file.*;
import java.util.*;
import java.util.stream.Stream;

public class WrapperClassLoader extends ClassLoader implements Closeable {
    private final Set<Path> sources = new LinkedHashSet<>();
    private final Set<Closeable> closeables = new HashSet<>();
    private final ClassLoader parent;

    static {
        registerAsParallelCapable();
    }

    public WrapperClassLoader(Iterable<Path> jars, ClassLoader parent) throws IOException, URISyntaxException {
        super(null);
        this.parent = parent;
        for (Path jar : jars) {
            if (Files.isDirectory(jar)) {
                sources.add(jar);
            } else {
                FileSystem fs = JFiles.openZipFile(jar, false, parent);
                closeables.add(fs);
                sources.add(fs.getPath(""));
            }
        }
    }

    @Override
    protected Class<?> findClass(String name) throws ClassNotFoundException {
        Path path = findClassSource(name);
        if (path == null) return parent.loadClass(name);
        byte[] clazz;
        try {
            clazz = Files.readAllBytes(path);
        } catch (IOException e) {
            throw new ClassNotFoundException("Could not find class", e);
        }
        return defineClass(name, clazz, 0, clazz.length);
    }

    @Override
    protected Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {
        if (findClassSource(name) == null || name.equals("io.gitlab.jfronny.inceptum.WrapperStrap"))
            return parent.loadClass(name);
        synchronized (this.getClassLoadingLock(name)) {
            Class<?> c = this.findLoadedClass(name);
            if (c == null) c = this.findClass(name);
            if (resolve) this.resolveClass(c);
            return c;
        }
    }

    @Override
    public URL getResource(String name) {
        Path p = findSource(name);
        if (p == null) return parent.getResource(name);
        return toUrl(p);
    }

    @Override
    protected Enumeration<URL> findResources(String name) {
        Stream<URL> resources = findPaths(name).map(this::toUrl);
        resources = Stream.concat(resources, parent.resources(name));
        return Collections.enumeration(resources.toList());
    }

    private URL toUrl(Path path) {
        try {
            return path.toUri().toURL();
        } catch (MalformedURLException e) {
            throw new RuntimeException("Could not get resource", e);
        }
    }

    private @Nullable Path findClassSource(String className) {
        return findSource(className.replace('.', '/') + ".class");
    }

    private @Nullable Path findSource(String path) {
        return findPaths(path).findFirst().orElse(null);
    }

    private Stream<Path> findPaths(String path) {
        return sources.stream().map(s -> s.resolve(path)).filter(Files::exists);
    }

    @Override
    public void close() throws IOException {
        for (Closeable closeable : closeables) {
            closeable.close();
        }
    }
}
